/**************************************************************************
                fb.c 0.4 Copyright (C) 2002, hito
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/fb.h>

#include "fb.h"

#define FB_ENV		"FRAMEBUFFER"
#define	FB_DEFDEV	"/dev/fb0"

#define MONO_OFFSET_8BIT  0x40
#define COLORS_MONO_8BIT  0x40
#define MONO_MASK_8BIT    0xFC
#define MONO_SHIFT_8BIT   2

#define COLOR_OFFSET_8BIT 0x80
#define COLORS_8BIT       0x80
#define RED_MASK_8BIT     0xC0
#define GREEN_MASK_8BIT   0xE0
#define BLUE_MASK_8BIT    0xC0
#define RED_SHIFT_8BIT    1
#define GREEN_SHIFT_8BIT  3
#define BLUE_SHIFT_8BIT   6

#define FALSE 0
#define TRUE  1

#define IMAGE_SIZE_MAX 10000

static struct fb_cmap *fb_cmap_create(struct fb_fix_screeninfo *,
				      struct fb_var_screeninfo *);
static void fb_cmap_destroy(struct fb_cmap *cmap);
static int fb_fscrn_get(int fbfp, struct fb_fix_screeninfo *scinfo);
static void *fb_mmap(int fbfp, struct fb_fix_screeninfo *scinfo);
static int fb_munmap(void *buf, struct fb_fix_screeninfo *scinfo);
static int fb_vscrn_get(int fbfp, struct fb_var_screeninfo *scinfo);
static int fb_cmap_set(int fbfp, struct fb_cmap *cmap);
static int fb_cmap_get(int fbfp, struct fb_cmap *cmap);

static struct fb_fix_screeninfo fscinfo;
static struct fb_var_screeninfo vscinfo;
static struct fb_cmap *cmap = NULL, *cmap_org = NULL;
static int is_open = FALSE;
static int fbfp = -1;
static size_t pixel_size = 0;
static unsigned char *buf = NULL;

int
fb_open(void)
{
    char *fbdev = { FB_DEFDEV };

    if (is_open == TRUE)
	return 1;

    if (getenv(FB_ENV)) {
	fbdev = getenv(FB_ENV);
    }

    if ((fbfp = open(fbdev, O_RDWR)) == -1) {
	fprintf(stderr, "open %s error\n", fbdev);
	goto ERR_END;
    }

    if (fb_fscrn_get(fbfp, &fscinfo)) {
	goto ERR_END;
    }

    if (fb_vscrn_get(fbfp, &vscinfo)) {
	goto ERR_END;
    }

    if ((cmap = fb_cmap_create(&fscinfo, &vscinfo)) == (struct fb_cmap *)-1) {
	goto ERR_END;
    }

    if (!(buf = fb_mmap(fbfp, &fscinfo))) {
	fprintf(stderr, "Can't allocate memory.\n");
	goto ERR_END;
    }

    if (fscinfo.type != FB_TYPE_PACKED_PIXELS) {
	fprintf(stderr, "This type of framebuffer is not supported.\n");
	goto ERR_END;
    }

    if (fscinfo.visual == FB_VISUAL_PSEUDOCOLOR && vscinfo.bits_per_pixel == 8) {
	if (fb_cmap_get(fbfp, cmap)) {
	    fprintf(stderr, "Can't get color map.\n");
	    fb_cmap_destroy(cmap);
	    cmap = NULL;
	    goto ERR_END;
	}

	pixel_size = 1;
    }
    else if ((fscinfo.visual == FB_VISUAL_TRUECOLOR ||
	      fscinfo.visual == FB_VISUAL_DIRECTCOLOR) &&
	     (vscinfo.bits_per_pixel == 15 ||
	      vscinfo.bits_per_pixel == 16 ||
	      vscinfo.bits_per_pixel == 24 || vscinfo.bits_per_pixel == 32)) {
	pixel_size = (vscinfo.bits_per_pixel + 7) / CHAR_BIT;
    }
    else {
	fprintf(stderr, "This type of framebuffer is not supported.\n");
	goto ERR_END;
    }

    is_open = TRUE;
    return 0;

  ERR_END:
    fb_close();
    return 1;
}

void
fb_close(void)
{
    if (is_open != TRUE)
	return;

    if (cmap != NULL) {
	fb_cmap_destroy(cmap);
	cmap = NULL;
    }
    if (cmap_org != NULL) {
	fb_cmap_set(fbfp, cmap_org);
	fb_cmap_destroy(cmap_org);
	cmap = NULL;
    }
    if (buf != NULL) {
	fb_munmap(buf, &fscinfo);
	buf = NULL;
    }

    if (fbfp >= 0) {
	close(fbfp);
    }

    is_open = FALSE;
}

unsigned char *fb_get_screen(void)
{
  unsigned char *tmp;

  if (is_open != TRUE)
    return NULL;

  tmp = malloc(vscinfo.yres * vscinfo.xres * pixel_size);
  if (tmp == NULL)
    return NULL;

  memcpy(tmp, buf, vscinfo.yres * vscinfo.xres * pixel_size);
  return tmp;
}

int
fb_buf_color(unsigned char *tmp_buf, int x, int y, int *r, int *g, int *b)
{
    unsigned long work = 0;
    int offset;

    if (tmp_buf == NULL || x >= vscinfo.xres || y >= vscinfo.yres)
	return 1;

    offset = fscinfo.line_length * y + pixel_size * x;

    if (offset >= fscinfo.smem_len)
	return 1;

    memcpy(&work, tmp_buf + offset, pixel_size);

    if (pixel_size == 1) {
	if (cmap == NULL)
	    return 1;
	if (cmap->red)
	    *r = *(cmap->red + work) >> CHAR_BIT;
	if (cmap->green)
	    *g = *(cmap->green + work) >> CHAR_BIT;
	if (cmap->blue)
	    *b = *(cmap->blue + work) >> CHAR_BIT;
    } else {
	*r = ((work >> vscinfo.red.
	       offset) & (0x000000ff >> (CHAR_BIT - vscinfo.red.length)))
	    << (CHAR_BIT - vscinfo.red.length);
	*g = ((work >> vscinfo.green.
	       offset) & (0x000000ff >> (CHAR_BIT - vscinfo.green.length)))
	    << (CHAR_BIT - vscinfo.green.length);
	*b = ((work >> vscinfo.blue.
	       offset) & (0x000000ff >> (CHAR_BIT - vscinfo.blue.length)))
	    << (CHAR_BIT - vscinfo.blue.length);
    }
    return 0;
}

void
fb_clear(void)
{
    if (is_open != TRUE)
	return;

    memset(buf, 0,
	   (vscinfo.xres * vscinfo.yres * vscinfo.bits_per_pixel) / CHAR_BIT);
}

int
fb_width(void)
{
    if (is_open != TRUE)
	return 0;

    return vscinfo.xres;
}

int
fb_height(void)
{
    if (is_open != TRUE)
	return 0;

    return vscinfo.yres;
}


/********* static functions **************/
/*
 * (struct fb_cmap)ǥХ˰¸ʤ顼ޥå׾
 * 
 * fb_cmap_create()     Υ顼ޥå׾
 * fb_cmap_destroy()    顼ޥå׾˴
 * fb_cmap_disp()               ɽ
 * fb_cmap_get()                γ
 * fb_cmap_set()                
 */

#define	LUT_MAX		(256)

static struct fb_cmap *
fb_cmap_create(struct fb_fix_screeninfo *fscinfo,
	       struct fb_var_screeninfo *vscinfo)
{
    struct fb_cmap *cmap;
    int cmaplen = LUT_MAX;

    /* 顼ޥåפ¸ߥå */
    if (fscinfo->visual == FB_VISUAL_MONO01 ||
	fscinfo->visual == FB_VISUAL_MONO10 ||
	fscinfo->visual == FB_VISUAL_TRUECOLOR)
	return NULL;

    cmap = (struct fb_cmap *)malloc(sizeof(struct fb_cmap));
    if (!cmap) {
	perror("cmap malloc error\n");
	return (struct fb_cmap *)-1;
    }
    memset(cmap, 0, sizeof(struct fb_cmap));

    /* ʬ¸ߤä饫顼ޥåѤΰ */
    if (vscinfo->red.length) {
	cmap->red = (__u16 *) malloc(sizeof(__u16) * cmaplen);
	if (!cmap->red) {
	    perror("red lut malloc error\n");
	    return (struct fb_cmap *)-1;
	}
    }
    if (vscinfo->green.length) {
	cmap->green = (__u16 *) malloc(sizeof(__u16) * cmaplen);
	if (!cmap->green) {
	    if (vscinfo->red.length)
		free(cmap->red);
	    perror("green lut malloc error\n");
	    return (struct fb_cmap *)-1;
	}
    }
    if (vscinfo->blue.length) {
	cmap->blue = (__u16 *) malloc(sizeof(__u16) * cmaplen);
	if (!cmap->blue) {
	    if (vscinfo->red.length)
		free(cmap->red);
	    if (vscinfo->green.length)
		free(cmap->green);
	    perror("blue lut malloc error\n");
	    return (struct fb_cmap *)-1;
	}
    }
    if (vscinfo->transp.length) {
	cmap->transp = (__u16 *) malloc(sizeof(__u16) * cmaplen);
	if (!cmap->transp) {
	    if (vscinfo->red.length)
		free(cmap->red);
	    if (vscinfo->green.length)
		free(cmap->green);
	    if (vscinfo->blue.length)
		free(cmap->blue);
	    perror("transp lut malloc error\n");
	    return (struct fb_cmap *)-1;
	}
    }
    cmap->len = cmaplen;
    return cmap;
}

static void
fb_cmap_destroy(struct fb_cmap *cmap)
{
    if (cmap->red)
	free(cmap->red);
    if (cmap->green)
	free(cmap->green);
    if (cmap->blue)
	free(cmap->blue);
    if (cmap->transp)
	free(cmap->transp);
    free(cmap);
}

static int
fb_cmap_get(int fbfp, struct fb_cmap *cmap)
{
    if (ioctl(fbfp, FBIOGETCMAP, cmap)) {
	perror("ioctl FBIOGETCMAP error\n");
	return -1;
    }
    return 0;
}

static int
fb_cmap_set(int fbfp, struct fb_cmap *cmap)
{
    if (ioctl(fbfp, FBIOPUTCMAP, cmap)) {
	perror("ioctl FBIOPUTCMAP error\n");
	return -1;
    }
    return 0;
}

/*
 * ե졼ХåեФ륢
 * 
 * fb_mmap()            ե졼ХåեΥؤΥޥå
 * fb_munmap()          ե졼ХåեΥ夫Υޥå
 */

static void *
fb_mmap(int fbfp, struct fb_fix_screeninfo *scinfo)
{
    void *buf;
    if ((buf = (unsigned char *)
	 mmap(NULL, scinfo->smem_len, PROT_READ | PROT_WRITE, MAP_SHARED, fbfp,
	      (off_t) 0))
	== MAP_FAILED) {
	perror("mmap error");
	return NULL;
    }
    return buf;
}

static int
fb_munmap(void *buf, struct fb_fix_screeninfo *scinfo)
{
    return munmap(buf, scinfo->smem_len);
}

/*
 * (struct fb_fix_screeninfo)ǥХ˰¸ʤꤵ줿
 * 
 * fb_fscrn_disp()              ɽ
 * fb_fscrn_get()               γ
 */
static int
fb_fscrn_get(int fbfp, struct fb_fix_screeninfo *scinfo)
{
    if (ioctl(fbfp, FBIOGET_FSCREENINFO, scinfo)) {
	perror("ioctl FBIOGET_FSCREENINFO error\n");
	return -1;
    }
    return 0;
}

/*
 * (struct fb_var_screeninfo)ǥХ˰¸ʤѹǽʾ
 * 
 * fb_vscrn_disp()              ɽ
 * fb_vscrn_get()               γ
 * fb_vscrn_set()               
 */
static int
fb_vscrn_get(int fbfp, struct fb_var_screeninfo *scinfo)
{
    if (ioctl(fbfp, FBIOGET_VSCREENINFO, scinfo)) {
	perror("ioctl FBIOGET_VSCREENINFO error\n");
	return -1;
    }
    return 0;
}
